import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import pytest
import sys
from pathlib import Path

import pkg_resources

main_folder = Path(__file__).parent.parent
sys.path.insert(0, str(main_folder))
sys.path.insert(0, str(main_folder / 'jetdnn'))
sys.path.insert(0, str(main_folder / 'tests'))
os.chdir(main_folder / 'jetdnn')

import jetdnn

"""
Assumes that test_predict is successful.
"""

def test_get_equation():

    csv = "table_EUROfusion_db_JSimpson_24april2019_D_withpellets_normp_nokikcs_only_validated.dat"
    csv_path = os.path.abspath("../" + csv)

    input_cols = ["Ip (MA)","P_tot (MW)","B (T)"]
    real_output_col = ["Te ped height pre-ELM (keV)"]
    test_data = jetdnn.predict.build_and_test_single(csv_path,input_cols,real_output_col)[3]
    testinput = test_data[input_cols]

    model, output_expected, error, data = jetdnn.predict.build_and_test_single(csv_path,input_cols,real_output_col)

    output_expected = np.array(output_expected)
    output = np.array(jetdnn.visualise.get_equation(model,data,input_cols,real_output_col)[2]).flatten()

    #print(jetdnn.visualise.get_equation(model,csv_path,input_cols,real_output_col)[0])

    assert output == pytest.approx(output_expected,abs=0.8) # checks that the output of the neural network checks out with the tensorflow predicted values


import numpy as np
import pandas as pd
import tensorflow as tf
import matplotlib.pyplot as plt
import os
import seaborn as sns

from tensorflow import keras
from keras import layers

print(tf.__version__)

def get_train_test_data(dataset):

    """
    get_train_test_data

    Split a chosen dataset randomly into an 80/20 train-test split. Display Numpy shapes of train and test data.

    Args:
        dataset (string): filename of csv.
    Returns:
        array: training data.
        array: test data.
    """

    train_data, test_data = np.split(dataset.sample(frac=1,random_state=42),[int(.8*len(dataset))]) # this works!
    print("Train data shape is: " + str(np.shape(train_data)))
    print("Test data shape is: " + str(np.shape(test_data)))
    return train_data, test_data

"""
def read_data(csv):
    data = pd.read_csv(csv,sep="\s{3,}|\s{3,}|\t+|\s{3,}\t+|\t+\s{3,}",skipinitialspace=True) # maybe change separator depending on testing
    print("Column names: " + str(data.columns)) # suggest doing this first to check the column names for input
    return(data)
"""

def build_and_test_single(csv_name,input_cols,output_col,plot_col=None,learning_rate=None,epochs=None,batch_size=None,maxoutput=None): # arguments set to None are optional

    """
    build_and_test_single

    Build and test a model predicting a single column of data from multiple input columns. Calculate the mean squared error of model predictions.

    Args:
        csv_name (string): filename of csv.
        input_cols (list): list of column names (as strings) of engineering parameters to predict from.
        output_col (string): name of column you'd like to predict.
        plot_col (string): optional. Name of column you'd like to plot your predictions against, selected from the engineering parameters.
        learning_rate (integer): optional. Value from 0.1 - 0.0001 representing the learning rate of the Keras Adam optimizer.
            Documentation for the Adam optimizer can be found here: https://keras.io/api/optimizers/adam/.
        epochs (integer): optional. Number of epochs for training the DNN. Typical starting values may be 50 for a smaller dataset up to 500 for a large dataset.
            If not given, the function will use a default value of 50.
        batch_size (integer): optional. Batch size for training the DNN, either 8, 16, 32 or 64. Smaller batch sizes may give more accurate predictions but take more time.
            If not given, the function will use a default value of 8.
        maxoutput (Boolean): optional. Set equal to True to display a more detailed output, useful for debugging.

    Returns:
        array: pedestal height predictions.
        integer: mean squared error of these predictions.
        Tensorflow Model: trained DNN model.

    """

    data = pd.read_csv(csv_name,sep="\s{3,}|\s{3,}|\t+|\s{3,}\t+|\t+\s{3,}",skipinitialspace=True)

    # search column names for each listed in input_cols and output_col
    # if none listed, add an error message

    """
    SPLIT TRAIN AND TEST DATA
    """

    train_data, test_data = get_train_test_data(data)
    if maxoutput == True:
        print("Train and test data are of type " + str(type(train_data)))
        print("Train data size is " + str(train_data.size))
        print("Test data size is " + str(test_data.size))
    
    """
    GET INPUTS AND OUTPUTS
    """

    traininput = train_data[input_cols] # this is the correct format, list of a list, or [["x","y"]]
    trainoutput = train_data[output_col]

    testinput = test_data[input_cols]
    testoutput = test_data[output_col]

    """
    NORMALISATION LAYER
    """
    normalizer = tf.keras.layers.Normalization(axis=-1)
    normalizer.adapt(np.array(traininput)) # adapting traininput means the DNN will accept however many input columns are given

    """
    MODEL BUILDING
    """

    def build_and_compile_model(norm,learnrate=None):
        model = keras.Sequential([
            norm,
            layers.Dense(128,activation="relu"),
            #layers.Dropout(0.2), # could experiment with removing dropout layers, e.g. a separate function
            layers.Dense(128,activation="relu"),
            #layers.Dropout(0.2),
            layers.Dense(1)
        ])
        if learnrate==None:
            model.compile(loss="mean_absolute_error",optimizer=tf.keras.optimizers.Adam(0.0005))
        else:
            model.compile(loss="mean_absolute_error",optimizer=tf.keras.optimizers.Adam(learnrate)) # this part may be finicky
        return model

    dnn_model = build_and_compile_model(normalizer,learning_rate)
    dnn_model.summary()

    """
    TRAINING
    """
    if epochs == None and batch_size == None:
        history = dnn_model.fit(x=traininput,y=trainoutput,validation_data=(testinput,testoutput),verbose=0,epochs=50,batch_size=8)
    elif epochs != None and batch_size == None:
        history = dnn_model.fit(x=traininput,y=trainoutput,validation_data=(testinput,testoutput),verbose=0,epochs=epochs,batch_size=8)
    elif epochs == None and batch_size != None:
        history = dnn_model.fit(x=traininput,y=trainoutput,validation_data=(testinput,testoutput),verbose=0,epochs=50,batch_size=batch_size)
    else:
        history = dnn_model.fit(x=traininput,y=trainoutput,validation_data=(testinput,testoutput),verbose=0,epochs=epochs,batch_size=batch_size) # options given if optional arguments are input


    """
    PREDICTION
    """
    predicted_ped = dnn_model.predict(testinput)

    """
    COMPILING
    The loss finds the mean error, then the optimizer aims to minimise this by changing the weights.
    """
    dnn_model.summary()

    def plot_loss(history):
        plt.plot(history.history["loss"], label="loss")
        plt.plot(history.history["val_loss"], label="val_loss")
        plt.ylim([0,5])
        plt.xlabel("Epoch")
        plt.ylabel(r"Error [pedestal height units]") # get this to obtain units from dataset
        plt.legend()
        plt.grid(True)

    plot_loss(history)
    plt.show()

    """
    RESHAPE DATA FOR PLOTTING
    """

    if maxoutput == True:
        print(np.shape(predicted_ped))

    flat_ped = []
    for sublist in predicted_ped.tolist():
        for item in sublist:
            flat_ped.append(item)
    
    if maxoutput == True:
        print(len(flat_ped))
        print(len(testinput))

    """
    PLOTTING

    Plots predicted and true pedestal heights against a chosen column from the input values.
    """

    if plot_col != None:
        x = list(testinput[plot_col].values.flatten())
    else:
        x = list(testinput[input_cols[0]].values.flatten())

    testoutput = list(testoutput.values.flatten())
    sns.relplot(x=x,y=flat_ped,alpha=.5,color="hotpink",label="Predicted values")
    sns.relplot(x=x,y=testoutput,alpha=.5,color="blue",label="True values")
    plt.xlabel(plot_col)
    plt.ylabel(output_col)
    plt.legend()
    plt.show()

    """
    PLOT ACCURACY
    """

    x = np.linspace(min(np.array(testoutput)),max(np.array(testoutput)),100)
    y = x
    sns.relplot(x=testoutput,y=flat_ped,alpha=.5,color="blueviolet",label = "Data points")
    plt.plot(x,y,color="blueviolet",label="y = x")
    plt.title("Accuracy")
    plt.xlabel(r"True values [$m^{-3}$]")
    plt.ylabel(r"Predicted values [$m^{-3}$]")
    plt.legend()
    plt.show()

    """
    EVALUATE ACCURACY
    """

    errs = []

    for n in range(0,len(flat_ped)):
        err = (flat_ped[n] - np.array(testoutput)[n])**2
        errs.append(err)

    m_s_e = sum(errs) / len(errs)
    print(m_s_e)

    # ensures 1d list output

    return dnn_model, flat_ped, m_s_e, test_data # model, predictions and error

def predict_single(model,filename,input_cols):
    """
    predict_single

    Predict pedestal heights using a trained DNN created with build_and_test_single.

    Args:
        model (Tensorflow Model): trained DNN from build_and_test_single to predict a column with.
        input_cols (array): input columns to apply the trained model on.

    Returns:
        array: predicted pedestal heights.
    """
    df = pd.read_csv(filename,sep="\s{3,}|\s{3,}|\t+|\s{3,}\t+|\t+\s{3,}",skipinitialspace=True)
    predict_data = df[input_cols] # this must be same size as train_data[input_cols], refer to output of previous function
    predictions = model.predict(predict_data)

    #predictions = list(predictions.flatten())

    return predictions



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import pytest
import sys
from pathlib import Path

import pkg_resources

main_folder = Path(__file__).parent.parent
sys.path.insert(0, str(main_folder))
sys.path.insert(0, str(main_folder / 'jetdnn'))
sys.path.insert(0, str(main_folder / 'tests'))
os.chdir(main_folder / 'jetdnn')

import jetdnn

# DNN TESTING FUNCTIONS
"""
The error provided is the error found for temperature data in the dataset used for "On neutral ground: can neutrals influence plasma pedestal formation?" by S. Frankel, J. Simpson and E. Solano
It is suggested to use this dataset or temperature pedestal data for testing with the below functions.

Assumes that test_data_manip is successful.
"""

test_csv = "table_EUROfusion_db_JSimpson_24april2019_D_withpellets_normp_nokikcs_only_validated.dat"
df = pd.read_csv(os.path.abspath("../" + test_csv),sep="\s{3,}|\s{3,}|\t+|\s{3,}\t+|\t+\s{3,}",skipinitialspace=True)
print(df.columns)

def test_build_and_test_single():

    csv = "table_EUROfusion_db_JSimpson_24april2019_D_withpellets_normp_nokikcs_only_validated.dat"
    csv_path = os.path.abspath("../" + csv)

    input_cols = ["Ip (MA)","P_tot (MW)","B (T)"]
    real_output_col = ["Te ped height pre-ELM (keV)"]

    test_data = jetdnn.predict.build_and_test_single(csv_path,input_cols,real_output_col)[3]
    output_expected = test_data[real_output_col].values.flatten().tolist()
    output = jetdnn.predict.build_and_test_single(csv_path,input_cols,real_output_col)[1] # returns flat_ped, or the predictions

    assert output == pytest.approx(output_expected,abs=0.7) # 0.7 found to be the absolute error with testing

def test_predict_single():
    
    csv = "table_EUROfusion_db_JSimpson_24april2019_D_withpellets_normp_nokikcs_only_validated.dat" # can be .csv or .dat format
    csv_path = os.path.abspath("../" + csv)
    df = pd.read_csv(csv_path,sep="\s{3,}|\s{3,}|\t+|\s{3,}\t+|\t+\s{3,}",skipinitialspace=True)

    input_cols = ["Ip (MA)","P_tot (MW)","B (T)"]
    real_output_col = ["Te ped height pre-ELM (keV)"]

    test_model = jetdnn.predict.build_and_test_single(csv_path,input_cols,real_output_col)[0]
    output = jetdnn.predict.predict_single(test_model,csv_path,input_cols)

    output_expected = df[real_output_col].values.flatten().tolist()

    assert output == pytest.approx(output_expected,abs=0.7) # 0.7 found to be the absolute error with testing


import seaborn as sns
import numpy as np
from jetdnn.predict import get_train_test_data # using just .predict gives an error

def plot_params(filename,plot_cols):

    """
    plot_params

    Create a Seaborn pairplot of data, and calculate the correlation coefficients of each subplot to find parameters to input in the DNN.
    Parameters with a strong correlation will be suggested for use in model building.

    Arguments:
        filename (string): name of .dat or .csv file.
        plot_cols (list): list of column names to plot and analyse.

    Returns:
        array: list of columns which are strongly correlated with each other.
    """

    plotting_data = get_train_test_data(filename)
    g = sns.pairplot(plotting_data[plot_cols],diag_kind = "kde") # plots parameters

    labels = []
    print("Correlation significant between:")

    for ax in g.axes:
        for curve in ax.lines:
            
            r = np.corrcoef(curve.get_xdata(),curve.get_ydata()) # find correlation matrix for each subplot

            if (r != 1) and ((r > 0.5) or (r < -0.5)): # if significant correlation
                print(curve.get_label) # list x and y labels
                labels.append(curve.get_label())
            else:
                print() # do nothing

    return labels



import tensorflow
from tensorflow import keras
import predict
from keras.utils import plot_model
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import re
from sympy import *

# get equation

import numpy as np
import tensorflow as tf

def get_equation(model,test_data,input_cols,output_col):
    """
    get_equation

    Produce an analytic equation from a trained DNN model, using forward propagation with numerical methods.

    Args:
        model (Tensorflow Model): trained DNN from build_and_test_single.
        input_cols (list): list of input columns as strings (should be the same as those used to test the DNN).

    Returns:
        string: short_eqn, an analytic equation representing the DNN model.
        list: x, the predicted values (should align with those returned by model.predict).
    """

    def activation_relu(input):
        
        output = np.maximum(0,input)

        return output
        
    node_equations = []

    #data = pd.read_csv(filename,sep="\s{3,}|\s{3,}|\t+|\s{3,}\t+|\t+\s{3,}",skipinitialspace=True)

    x = test_data[input_cols]
    layer_nodes = [len(input_cols),128,128,1]
    
    for n in range(0,len(model.layers)): # for each layer
        layer = model.layers[n]
        weights = layer.get_weights() # get list of weights for one layer e.g. W[0][1],W[0][2]
        b = weights[1] # define bias
        w = weights[0] # define weights
        nodes = []

        if n != 0: # try removing this next time you're testing

            for j in range(0,layer_nodes[n]): # for each node (e.g. for each weight in w0 e.g. w01, w02, w03 etc)
                node = 0
                node_eqn = ""

                for i in range(0,len(input_cols)): # for each input, add weights and inputs to node

                    if n == 1: # why 1 and not 0??
                        x_eqn = "x" + str(i)
                    
                    if isinstance(x,pd.DataFrame):
                        col = np.array(x.iloc[:,i]) # gets ith column, converts to numpy array
                    else:
                        col = x[i]

                    # BUILD NODE
                    
                    node += (col * float(w[i,j])) # gives the jth value in i

                    # BUILD MAIN NODE EQUATION BEFORE ADDING BIAS
                    
                    if x_eqn != "x" + str(i): # if not the first layer: IGNORE FOR FIRST PASS OF n!
                        if node_equations[i] == node_equations[-1]: # if last iteration
                            x_eqn = node_equations[i]
                            node_eqn += "(" + str(x_eqn) + ")" + "*" + str(w[i,j]) # finish off equation
                        else:
                            x_eqn = node_equations[i] # multiply appropriate equation with weight
                            node_eqn += "(" + str(x_eqn) + ")" + "*" + str(w[i,j]) + " + "

                    else:
                        if i == len(input_cols): # if the final iteration, finishes off expression
                            node_eqn += str(x_eqn) + "*" + str(w[i,j])
                        else:
                            node_eqn += str(x_eqn) + "*" + str(w[i,j]) + " + "

                    # BUILDS NODES AND NODE EQUATIONS (WITHOUT BIASES) BY END OF ALL i ITERATIONS
                
                node += b[j] # node now contains weights and bias
                node_eqn += str(b[j])
                
                if activation_relu(node) is not 0: # if +ive node value

                    nodes.append(activation_relu(node)) # add node to node list for that layer
                    node_equations.append(node_eqn) # add node equation to list for that layer


            x = nodes # node values become new inputs
            x_eqn = node_equations # list of node equations for that layer
                
    

    # now out of all loops
    
    final_eqn = node_equations[-1] # this should be the final equation

    smpl = simplify(final_eqn)

    print(final_eqn)
    print(smpl)
    print(x)

    #x = list(x.flatten()) # converts x to 1d list (not needed, x already list)

    return smpl, final_eqn, x


# VISUALISE MODELS

def display_model(model):

    """
    display_model

    Display a summary and save a Keras plot of a Deep Neural Network model.

    Args:
        model (Tensorflow Model): Deep Neural Network model build with build_and_test_single.
    
    Returns:
        None
    """
    print(model.summary())
    plot_model(model, to_file='dnn_model_plot.png', show_shapes=True, show_layer_names=True)
    return

# density npm plot



from setuptools import setup, find_packages

setup(
    name = "JETDNN",
    version = "1.3.1",
    author = "S. Frankel, J. Simpson and E. Solano.",
    author_email = "s.frankel2@newcastle.ac.uk",
    url = "https://github.com/quasoph/JETDNN",
    description = "JETDNN: a Python package using Deep Neural Networks (DNNs) to find H-mode plasma pedestal heights from multiple engineering parameters.",
    packages = find_packages()
)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import pytest
import sys
from pathlib import Path

import pkg_resources

main_folder = Path(__file__).parent.parent
sys.path.insert(0, str(main_folder))
sys.path.insert(0, str(main_folder / 'jetdnn'))
sys.path.insert(0, str(main_folder / 'tests'))
os.chdir(main_folder / 'jetdnn')

import jetdnn
g = pkg_resources.get_distribution("jetdnn").version
print("jetdnn version: " + str(g))

# DATA MANIPULATION TESTS


